#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/sched/signal.h>
#include <linux/pgtable.h>
#include <linux/moduleparam.h>
#include <linux/sched/mm.h>
#include <linux/mm.h>

MODULE_DESCRIPTION("LKM4");
MODULE_AUTHOR("CS695");
MODULE_LICENSE("GPL");

static int pid=1;
module_param(pid, int, 0);
MODULE_PARM_DESC(pid, "PID of the process");

unsigned long get_mapped_size(struct mm_struct* mm, struct vm_area_struct* vma) {
    unsigned long size = 0;
	pgd_t *pgd;
    p4d_t *p4d;
    pud_t *pud;
    pmd_t *pmd;
    pte_t *pte;
    
    for (unsigned long vaddr = vma->vm_start; vaddr < vma->vm_end; vaddr += PAGE_SIZE) {
        if (vaddr & (PAGE_SIZE - 1))
            pr_info("[LKM4] Error in iteration\n");
        
		// Page table walk
		while (1) {
			pgd = pgd_offset(mm, vaddr);
			if (pgd_none(*pgd) || pgd_bad(*pgd))
				break;

			p4d = p4d_offset(pgd, vaddr);
			if (p4d_none(*p4d) || p4d_bad(*p4d))
				break;

			pud = pud_offset(p4d, vaddr);
			if(pud_none(*pud) || pud_bad(*pud)) 
				break;

			pmd = pmd_offset(pud, vaddr);
			if(pmd_none(*pmd) || pmd_bad(*pmd))
				break;

			pte = pte_offset_kernel(pmd, vaddr);
			if(pte) {
				pte_unmap(*pte);
				
				if (pte_present(*pte))
					size += PAGE_SIZE;
			}
			break;
		}
    }
    return size;
}

static int __init print_alloc_size(void)
{
	struct task_struct *task;
	struct mm_struct *mm;
	unsigned long virtual_size = 0;
	unsigned long physical_size = 0;

	task = pid_task(find_vpid(pid), PIDTYPE_PID);
	if (task == NULL) {
		pr_info("[LKM4] Process with PID: %d not found\n", pid);
		return 0;
	}
	mm = task -> mm;
	if (mm != NULL) {
		/* Virtual Size */
		virtual_size = (mm -> total_vm << PAGE_SHIFT) / 1024;

		/* Alternative iterative way to fine virtual size */
		// struct vm_area_struct *vma;
		// VMA_ITERATOR(iter, mm, 0);

        // for_each_vma(iter, vma) {
        //     virtual_size += (vma->vm_end - vma->vm_start);
        // }

		// virtual_size /= 1024;

		pr_info("[LKM4] Virtual Memory Size: %lu KiB\n", virtual_size);

		/* Physical Size (Not accurate (but fast) - Used by procfs status, statm - ps, top [RES value], etc.) */
		physical_size = (get_mm_rss(mm) << PAGE_SHIFT) / 1024;

		/* Alternative: Add up all page counts (this what get_mm_rss does internally) */
		// physical_size += atomic_long_read(&mm -> rss_stat.count[MM_FILEPAGES]);
		// physical_size += atomic_long_read(&mm -> rss_stat.count[MM_ANONPAGES]);
		// physical_size += atomic_long_read(&mm -> rss_stat.count[MM_SHMEMPAGES]);

		// physical_size = (physical_size << PAGE_SHIFT) / 1024;

		pr_info("[LKM4] Estimated Physical Memory Size: %lu KiB\n", physical_size);

		/* Physical Size (Very precise (but slow) - Used by procfs smaps, top [RSS value], etc.) 
		 * Ref: https://www.kernel.org/doc/html/latest/filesystems/proc.html?highlight=precise
		 *		https://www.kernel.org/doc/Documentation/ABI/testing/procfs-smaps_rollup
		 *		https://man7.org/linux/man-pages/man1/top.1.html?highlight=rss
		*/
		struct vm_area_struct *vma;
		VMA_ITERATOR(iter, mm, 0);
		physical_size = 0;

        for_each_vma(iter, vma) {
            physical_size += get_mapped_size(mm, vma);
        }

		physical_size /= 1024;

		pr_info("[LKM4] Precise Physical Memory Size: %lu KiB\n", physical_size);
	}
	else
		pr_info("[LKM4] mm is NULL\n");

	return 0;
}

static void __exit exit_module(void)
{
	pr_info("[LKM4] Module LKM4 Unloaded\n");
}

module_init(print_alloc_size);
module_exit(exit_module);
